from typing import TypeVar

from aio_microservice import Service, ServiceSettings
from aio_microservice.core.abc import ExtensionABC


async def test_custom_base() -> None:
    TestBaseSettingsT = TypeVar("TestBaseSettingsT", bound=ServiceSettings)

    class TestBase(Service[TestBaseSettingsT]): ...

    class TestSettings(ServiceSettings):
        somethings: str = "else"

    class TestService(TestBase[TestSettings]): ...

    service = TestService()
    assert service.settings.somethings == "else"


async def test_custom_base_with_extensions() -> None:
    class TestExtension(ExtensionABC): ...

    TestBaseSettingsT = TypeVar("TestBaseSettingsT", bound=ServiceSettings)

    class TestBase(Service[TestBaseSettingsT], TestExtension): ...

    class TestSettings(ServiceSettings):
        somethings: str = "else"

    class TestService(TestBase[TestSettings]): ...

    service = TestService()
    assert service.settings.somethings == "else"



from typing import TypeVar

from litestar import Litestar
from litestar.testing import AsyncTestClient

from aio_microservice.core.service import Service, ServiceSettingsT

ServiceT = TypeVar("ServiceT", bound=Service[ServiceSettingsT])  # type: ignore[valid-type]


class TestHttpClient(AsyncTestClient[Litestar]):
    def __init__(self, service: ServiceT) -> None:
        super().__init__(app=service._litestar_app)
        self._service = service



from __future__ import annotations

from aio_microservice import (
    Service,
    ServiceSettings,
    http,
    liveness_probe,
    readiness_probe,
)
from aio_microservice.http import TestHttpClient


async def test_readiness_probe() -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.test_readiness = True

        @readiness_probe
        async def test_readiness_probe(self) -> bool:
            return self.test_readiness

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/readiness")
        assert response.status_code == http.status_codes.HTTP_200_OK

        service.test_readiness = False
        response = await client.get("/readiness")
        assert response.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE


async def test_readiness_probe_stacking() -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.test_readiness_1 = True
            self.test_readiness_2 = True

        @readiness_probe
        async def test_readiness_probe_1(self) -> bool:
            return self.test_readiness_1

        @readiness_probe
        async def test_readiness_probe_2(self) -> bool:
            return self.test_readiness_2

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/readiness")
        assert response.status_code == http.status_codes.HTTP_200_OK

        service.test_readiness_2 = False
        response = await client.get("/readiness")
        assert response.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE


async def test_liveness_probe() -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.test_liveness = True

        @liveness_probe
        async def test_liveness_probe(self) -> bool:
            return self.test_liveness

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/liveness")
        assert response.status_code == http.status_codes.HTTP_200_OK

        service.test_liveness = False
        response = await client.get("/liveness")
        assert response.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE


async def test_liveness_probe_stacking() -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.test_liveness_1 = True
            self.test_liveness_2 = True

        @liveness_probe
        async def test_liveness_probe_1(self) -> bool:
            return self.test_liveness_1

        @liveness_probe
        async def test_liveness_probe_2(self) -> bool:
            return self.test_liveness_2

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/liveness")
        assert response.status_code == http.status_codes.HTTP_200_OK

        service.test_liveness_2 = False
        response = await client.get("/liveness")
        assert response.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE



from aio_microservice.http_cors.extension import (
    HttpCorsExtension,
    HttpCorsExtensionSettings,
    HttpCorsSettings,
)

__all__ = [
    "HttpCorsExtension",
    "HttpCorsExtensionSettings",
    "HttpCorsSettings",
]



from __future__ import annotations

from collections.abc import AsyncGenerator, AsyncIterator

import strawberry

from aio_microservice import Service, ServiceSettings
from aio_microservice.graphql import BaseContext, GraphqlContext, GraphqlExtension
from aio_microservice.http import RequestFactory


async def test_graphql_test_client_query() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(self) -> str:
            return "TEST RESPONSE"

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

    service = TestService()
    query = "query { test }"
    response = await service.graphql.schema.execute(query)
    assert response.errors is None
    assert response.data == {"test": "TEST RESPONSE"}


async def test_graphql_test_client_query_with_default_context() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(
            self,
            info: strawberry.types.Info[GraphqlContext[TestService], None],
        ) -> str:
            return info.context.service.response

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.response = "TEST RESPONSE"

    service = TestService()
    query = "query { test }"
    response = await service.graphql.schema.execute(query)

    assert response.errors is None
    assert response.data == {"test": "TEST RESPONSE"}


async def test_graphql_test_client_query_with_custom_context() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(
            self,
            info: strawberry.types.Info[GraphqlContext[TestService], None],
        ) -> str:
            custom_header = info.context.request.headers["custom-header"]  # type: ignore
            return f"{info.context.service.response}-{custom_header}"

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.response = "TEST"

    service = TestService()

    query = "query { test }"

    request = RequestFactory().get("/graphql", headers={"custom-header": "HEADER"})
    context_value = BaseContext(request=request)
    response = await service.graphql.schema.execute(query, context_value=context_value)

    assert response.errors is None
    assert response.data == {"test": "TEST-HEADER"}


async def test_graphql_test_client_subscription() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def dummy(self) -> bool:
            return True

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def test(self, target: int = 100) -> AsyncGenerator[int, None]:
            for i in range(target):
                yield i

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query, subscription=Subscription)

    service = TestService()
    query = "subscription { test(target: 3) }"
    subscription = await service.graphql.schema.subscribe(query)

    assert isinstance(subscription, AsyncIterator)
    responses = [x async for x in subscription]
    assert len(responses) == 3
    assert responses[0].data == {"test": 0}
    assert responses[1].data == {"test": 1}
    assert responses[2].data == {"test": 2}


async def test_graphql_test_client_subscription_with_default_context() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def dummy(self) -> bool:
            return True

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def test(
            self,
            info: strawberry.types.Info[GraphqlContext[TestService], None],
            target: int = 100,
        ) -> AsyncGenerator[str, None]:
            for i in range(target):
                yield f"{info.context.service.response_prefix}-{i}"  # type: ignore

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query, subscription=Subscription)

        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.response_prefix = "TEST"

    service = TestService()
    query = "subscription { test(target: 3) }"
    subscription = await service.graphql.schema.subscribe(query)

    assert isinstance(subscription, AsyncIterator)
    responses = [x async for x in subscription]
    assert len(responses) == 3
    assert responses[0].data == {"test": "TEST-0"}
    assert responses[1].data == {"test": "TEST-1"}
    assert responses[2].data == {"test": "TEST-2"}


async def test_graphql_test_client_subscription_with_custom_context() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def dummy(self) -> bool:
            return True

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def test(
            self,
            info: strawberry.types.Info[GraphqlContext[TestService], None],
            target: int = 100,
        ) -> AsyncGenerator[str, None]:
            custom_header = info.context.request.headers["custom-header"]  # type: ignore
            for i in range(target):
                yield f"{info.context.service.response_prefix}-{custom_header}-{i}"  # type: ignore

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query, subscription=Subscription)

        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.response_prefix = "TEST"

    service = TestService()
    query = "subscription { test(target: 3) }"
    request = RequestFactory().get("/graphql", headers={"custom-header": "HEADER"})
    context_value = BaseContext(request=request)
    subscription = await service.graphql.schema.subscribe(query, context_value=context_value)

    assert isinstance(subscription, AsyncIterator)
    responses = [x async for x in subscription]
    assert len(responses) == 3
    assert responses[0].data == {"test": "TEST-HEADER-0"}
    assert responses[1].data == {"test": "TEST-HEADER-1"}
    assert responses[2].data == {"test": "TEST-HEADER-2"}



from __future__ import annotations

import dataclasses
import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, ClassVar, TypeVar

from faststream import BaseMiddleware, FastStream
from faststream.broker.utils import default_filter
from faststream.rabbit import RabbitBroker, RabbitExchange, RabbitQueue
from faststream.security import SASLPlaintext
from loguru import logger
from pydantic import BaseModel, Field, SecretStr
from typing_extensions import Concatenate, ParamSpec

from aio_microservice.amqp.asyncapi import make_asyncapi_controller
from aio_microservice.core.abc import (
    ExtensionABC,
    litestar_on_app_init,
    readiness_probe,
    shutdown_hook,
    startup_hook,
    startup_message,
)
from aio_microservice.types import Port  # noqa: TCH001

if TYPE_CHECKING:
    from collections.abc import Iterable

    from aio_pika.abc import TimeoutType
    from faststream.broker.types import (
        Filter,
        PublisherMiddleware,
        SubscriberMiddleware,
    )
    from faststream.rabbit.message import RabbitMessage
    from faststream.rabbit.schemas import ReplyConfig
    from faststream.types import AnyDict
    from litestar.config.app import AppConfig


class AmqpSettings(BaseModel):
    host: str = Field(
        default="localhost",
        description="The ip for connecting to the broker.",
    )
    port: Port = Field(
        default=5672,
        description="The port for connecting to the broker.",
    )
    username: str = Field(
        default="guest",
        description="The username for authentication on the broker.",
    )
    password: SecretStr = Field(
        default=SecretStr("guest"),
        description="The password for authentication on the broker.",
    )
    prefetch_count: int | None = Field(
        default=None,
        description="The prefetch window size.",
    )
    timeout_graceful_shutdown: float | None = Field(
        default=None,
        description="The timeout for graceful shutdown (in seconds).",
    )


class AmqpExtensionImpl:
    def __init__(self, service: AmqpExtension, settings: AmqpSettings) -> None:
        self._service = service
        self._settings = settings
        self._faststream_rabbit_broker = self._create_faststream_broker(
            service=service,
        )
        self._faststream_app = self._create_faststream_app(
            service=service,
            broker=self._faststream_rabbit_broker,
        )
        self._asyncapi_controller = make_asyncapi_controller(
            app=self._faststream_app,
        )
        self._register_handlers()

    def _create_faststream_broker(self, service: AmqpExtension) -> RabbitBroker:
        security = SASLPlaintext(
            username=self._settings.username,
            password=self._settings.password.get_secret_value(),
        )
        return RabbitBroker(
            host=self._settings.host,
            port=self._settings.port,
            security=security,
            middlewares=service.__amqp_middlewares__,
            max_consumers=self._settings.prefetch_count,
            graceful_timeout=self._settings.timeout_graceful_shutdown,
        )

    def _create_faststream_app(self, service: AmqpExtension, broker: RabbitBroker) -> FastStream:
        return FastStream(
            broker=broker,
            title=service.__class__.__name__,
            version=service.__version__,
            description=service.__description__,
        )

    def _register_handlers(self) -> None:
        handler_methods = inspect.getmembers(
            object=self._service,
            predicate=lambda obj: hasattr(obj, AmqpDecorator.MARKER),
        )
        for handler_name, _ in handler_methods:
            handler = getattr(self._service, handler_name)
            handler_settings = getattr(handler, AmqpDecorator.MARKER)
            for handler_setting in handler_settings:
                if isinstance(handler_setting, subscriber):
                    subscriber_decorator = self._faststream_rabbit_broker.subscriber(
                        **dataclasses.asdict(handler_setting),
                    )
                    handler = subscriber_decorator(handler)
                elif isinstance(handler_setting, publisher):  # pragma: no branch
                    publisher_decorator = self._faststream_rabbit_broker.publisher(
                        **dataclasses.asdict(handler_setting),
                    )
                    handler = publisher_decorator(handler)

    @property
    def broker(self) -> RabbitBroker:
        return self._faststream_rabbit_broker


class AmqpExtensionSettings(BaseModel):
    amqp: AmqpSettings = AmqpSettings()


class AmqpExtension(ExtensionABC):
    __amqp_middlewares__: ClassVar[list[type[BaseMiddleware]]] = []

    def __init__(self, settings: AmqpExtensionSettings) -> None:
        self.amqp = AmqpExtensionImpl(service=self, settings=settings.amqp)

    @litestar_on_app_init
    def amqp_litestar_on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.route_handlers.append(self.amqp._asyncapi_controller)
        return app_config

    @startup_message
    async def _amqp_startup_message(self) -> str:
        return """
            # Amqp

            The following http-endpoints are available:
            - {scheme}://{host}:{port}/schema/asyncapi
            - {scheme}://{host}:{port}/schema/asyncapi.json
            - {scheme}://{host}:{port}/schema/asyncapi.yaml
        """

    @startup_hook
    async def _amqp_startup_hook(self) -> None:
        logger.info("Connecting to broker")
        await self.amqp._faststream_rabbit_broker.start()

    @shutdown_hook
    async def _amqp_shutdown_hook(self) -> None:
        logger.info("Disconnecting from broker")
        await self.amqp._faststream_rabbit_broker.close()

    @readiness_probe
    async def _amqp_readiness_probe(self) -> bool:
        if self.amqp._faststream_rabbit_broker._connection:
            # NOTE this might break.
            # faststream maintainer says, they will provide functionality like this.
            return self.amqp._faststream_rabbit_broker._connection.connected.is_set()
        return False  # pragma: no cover


AmqpExtensionT = TypeVar("AmqpExtensionT", bound=AmqpExtension)
P = ParamSpec("P")
R = TypeVar("R")


class AmqpDecorator:
    MARKER = "_amqp_decorator"


@dataclass
class subscriber(AmqpDecorator):  # noqa: N801
    queue: RabbitQueue | str
    exchange: RabbitExchange | str | None = None
    consume_args: AnyDict | None = None
    reply_config: ReplyConfig | None = None

    # broker arguments
    middlewares: Iterable[SubscriberMiddleware[RabbitMessage]] = ()
    filter: Filter[RabbitMessage] = default_filter
    retry: bool | int = False
    no_ack: bool = False
    no_reply: bool = False

    # AsyncAPI information
    title: str | None = None
    description: str | None = None
    include_in_schema: bool = True

    def __call__(
        self,
        fn: Callable[Concatenate[AmqpExtensionT, P], R],
    ) -> Callable[Concatenate[AmqpExtensionT, P], R]:
        handlers = getattr(fn, self.MARKER, [])
        handlers.append(self)
        setattr(fn, self.MARKER, handlers)
        return fn


@dataclass
class publisher(AmqpDecorator):  # noqa: N801
    queue: RabbitQueue | str = ""
    exchange: RabbitExchange | str | None = None
    routing_key: str = ""
    mandatory: bool = True
    immediate: bool = False
    timeout: TimeoutType = None
    persist: bool = False
    reply_to: str | None = None
    priority: int | None = None

    # specific
    middlewares: Iterable[PublisherMiddleware] = ()

    # AsyncAPI information
    title: str | None = None
    description: str | None = None
    include_in_schema: bool = True

    def __call__(
        self,
        fn: Callable[Concatenate[AmqpExtensionT, P], R],
    ) -> Callable[Concatenate[AmqpExtensionT, P], R]:
        handlers = getattr(fn, self.MARKER, [])
        handlers.append(self)
        setattr(fn, self.MARKER, handlers)
        return fn



from typing import Any

from litestar import Request

from aio_microservice import Service, ServiceSettings, http
from aio_microservice.http import TestHttpClient
from aio_microservice.http_cors import HttpCorsExtension, HttpCorsExtensionSettings

# TODO add some more tests and asserts to check if settings actually apply


async def test_http_cors() -> None:
    class TestServiceSettings(ServiceSettings, HttpCorsExtensionSettings): ...

    class TestService(Service[TestServiceSettings], HttpCorsExtension):
        @http.get(path="/test")
        async def get_test(self, request: Request[Any, Any, Any]) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.options(
            url="/test",
            headers={
                "Origin": "https://example.com",
            },
        )
        assert "Access-Control-Allow-Origin" in response.headers



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from faststream.rabbit import RabbitBroker, TestRabbitBroker
from typing_extensions import override

from aio_microservice.amqp import AmqpExtension

if TYPE_CHECKING:
    from faststream.rabbit.publisher.asyncapi import AsyncAPIPublisher

ServiceT = TypeVar("ServiceT", bound=AmqpExtension)


class AmqpBroker:
    def __init__(self, broker: RabbitBroker) -> None:
        self._faststream_rabbit_broker = broker

    def __getattr__(self, attr: str) -> Any:  # noqa: ANN401
        return getattr(self._faststream_rabbit_broker, attr)

    def __setattr__(self, attr: str, val: Any) -> None:  # noqa: ANN401
        if attr == "_faststream_rabbit_broker":
            object.__setattr__(self, attr, val)
        else:
            setattr(self._faststream_rabbit_broker, attr, val)  # pragma: no cover

    def get_published_messages(
        self,
        queue: str | None = None,
        exchange: str | None = None,
    ) -> list[Any]:
        messages: list[Any] = []
        publisher: AsyncAPIPublisher | None = None

        for p in self._faststream_rabbit_broker._publishers.values():  # pragma: no branch
            if queue is not None and p.queue.name != queue:
                continue  # pragma: no cover
            if exchange is not None and p.exchange.name != exchange:
                continue  # pragma: no cover
            publisher = p
            break

        if publisher is None or publisher.mock is None:
            return messages  # pragma: no cover

        for call in publisher.mock.call_args_list:
            args, _ = call
            messages.append(args[0])
        return messages

    def reset_published_messages(self) -> None:
        for publisher in self._faststream_rabbit_broker._publishers.values():
            if publisher.mock is None:
                continue  # pragma: no cover
            publisher.mock.reset_mock()


class TestAmqpBroker(TestRabbitBroker):
    def __init__(self, service: ServiceT, with_real: bool = False) -> None:
        super().__init__(broker=service.amqp._faststream_rabbit_broker, with_real=with_real)
        self._service = service

    @override
    async def __aenter__(self) -> AmqpBroker:  # type: ignore
        broker = await super().__aenter__()
        return AmqpBroker(broker)



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar.config.cors import CORSConfig
from pydantic import BaseModel, Field

from aio_microservice.core.abc import ExtensionABC, litestar_on_app_init

if TYPE_CHECKING:
    from litestar.config.app import AppConfig


class HttpCorsSettings(BaseModel):
    allow_origins: str = Field(
        default="*",
        description="Comma separated list of allowed origins.",
    )
    allow_methods: str = Field(
        default="*",
        description="Comma separated list of allowed methods.",
    )
    allow_headers: str = Field(
        default="*",
        description="Comma separated list of allowed headers.",
    )
    allow_credentials: bool = Field(
        default=False,
        description="Whether or not to set 'Access-Control-Allow-Credentials'.",
    )
    allow_origin_regex: str | None = Field(
        default=None,
        description="Regex of allowed origins.",
    )
    expose_headers: str = Field(
        default="",
        description="Comma separated list of headers set in the 'Access-Control-Expose-Headers'.",
    )
    max_age: int = Field(
        default=600,
        description="Response cache TTL in seconds.",
    )


class HttpCorsExtensionSettings(BaseModel):
    http_cors: HttpCorsSettings = HttpCorsSettings()


class HttpCorsExtension(ExtensionABC):
    def __init__(self, settings: HttpCorsExtensionSettings) -> None:
        self._litestar_cors_config = CORSConfig(
            allow_origins=settings.http_cors.allow_origins.split(","),
            allow_methods=settings.http_cors.allow_methods.split(","),  # type: ignore[arg-type]
            allow_headers=settings.http_cors.allow_headers.split(","),
            allow_credentials=settings.http_cors.allow_credentials,
            allow_origin_regex=settings.http_cors.allow_origin_regex,
            expose_headers=settings.http_cors.expose_headers.split(","),
            max_age=settings.http_cors.max_age,
        )

    @litestar_on_app_init
    def http_cors_litestar_on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.cors_config = self._litestar_cors_config
        return app_config



from __future__ import annotations

import strawberry

from aio_microservice import Service, ServiceSettings, http
from aio_microservice.graphql import GraphqlExtension
from aio_microservice.http import TestHttpClient


async def test_graphql_schema_download() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(self) -> str:
            return "TEST RESPONSE"

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

    service = TestService()
    async with TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/graphql.graphql")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "test: String!" in response.text



from strawberry.litestar import BaseContext

from aio_microservice.graphql.extension import GraphqlContext, GraphqlExtension

__all__ = [
    "BaseContext",
    "GraphqlContext",
    "GraphqlExtension",
]
